// These will be added programmatically.
// Uncomment them temporarily to check compilation.
//#define ALBEDO_TINT
//#define ALPHA_TEST
//#define ENVIRONMENT_MAP
//#define FOG
//#define LIGHT_MAP

attribute vec4 a_position;
attribute vec3 a_normal;
attribute vec4 a_texcoord;

uniform mat4 u_View;
uniform mat4 u_ViewProj;
uniform vec2 u_TexAnim;

#ifdef FOG
uniform float u_FogNearPlane;
uniform float u_FogRcpRange;
uniform float u_FogOpacity;
#endif

// material uniforms
uniform vec3 u_xAlbedoColour;

//instance uniforms
uniform vec3 u_xInstancePosition;
uniform vec4 u_xInstanceOrientation[3];
uniform vec4 u_xInstanceRGB_SH[3];

uniform vec3 u_xCameraUp;
uniform vec3 u_xCameraRight;
uniform vec3 u_xCameraForwards;

uniform bool u_bIsInteractive;
uniform bool u_bIsBillboard;

#ifdef ENVIRONMENT_MAP
uniform highp vec3 u_CameraPosition;
#endif
 
// outputs
varying vec4 v_VertexColour;
varying vec2 v_TexCoord;
#ifdef FOG
varying float v_Fog;
#endif
#ifdef LIGHT_MAP
varying vec2 v_LightMapUV;
#endif
#ifdef ENVIRONMENT_MAP
varying mediump vec3 v_Reflection;
#endif


vec3 RotateVectorIntoWorldSpace (vec3 xVector,
							     vec3 xTransform0,
							     vec3 xTransform1,
							     vec3 xTransform2)
{
	vec3 xRotatedVector;

	// Rotate the x, y and z components into tangent-space...
	xRotatedVector.x	= dot(xVector, xTransform0);
	xRotatedVector.y	= dot(xVector, xTransform1);
	xRotatedVector.z	= dot(xVector, xTransform2);
	
	xRotatedVector = normalize( xRotatedVector);

	// Return the rotated vector...
	return	(xRotatedVector);
}


vec3 SHLookup( const vec4 xNormal, const vec4 xRedSH, const vec4 xGreenSH, const vec4 xBlueSH )
{
	vec3 xColour;

	xColour.r = dot( xNormal, xRedSH );
	xColour.g = dot( xNormal, xGreenSH );
	xColour.b = dot( xNormal, xBlueSH );

	xColour = max( xColour, 0.0 );
	
	return xColour;
}

void main()
{    
	vec4 xPositionWorld;
	
   	xPositionWorld.x = dot( a_position.xyz, u_xInstanceOrientation[0].xyz);
    xPositionWorld.y = dot( a_position.xyz, u_xInstanceOrientation[1].xyz);
    xPositionWorld.z = dot( a_position.xyz, u_xInstanceOrientation[2].xyz);	
	xPositionWorld.w = 1.0;
	
	vec3 xNormal;

	// Animate vertex 
#ifndef LIGHT_MAP
	if ( u_bIsInteractive )
	{
		// Currently storing the weight per vertex
		float fWeight = a_position.w;
		
		// This is the price we pay for packing our bone data!
		vec3 xBone;
		xBone.x = u_xInstanceOrientation[0].w;
		xBone.y = u_xInstanceOrientation[1].w;
		xBone.z = u_xInstanceOrientation[2].w;
    
		// We must scale our anim bone by the scale of the instance.
	 	// TODO - is there a nice way of getting this? Possibly we can handle
		// it CPU side. Note this assumes one scale factor across XYZ, which
		// is the current limitation on animating/billboard instances.
		fWeight *= sqrt(dot( u_xInstanceOrientation[0].xyz, u_xInstanceOrientation[0].xyz ));

		// Again, this is constant per vertex
		float fOldPositionLength = sqrt( dot( xPositionWorld.xyz, xPositionWorld.xyz) );

		// Offset the vertex, but then make sure the distance from the origin
		// doesn't change.
		vec3 xBonedPosition = xPositionWorld.xyz + (fWeight * xBone);
		xPositionWorld.xyz = normalize( xBonedPosition ) * fOldPositionLength;
	}
	
	
	if ( u_bIsBillboard )
	{
		// TODO - be good to precalculate this scale constant
		float fScale = sqrt(dot( u_xInstanceOrientation[0].xyz, u_xInstanceOrientation[0].xyz ));
	
		xPositionWorld.xyz += a_normal.x * fScale * u_xCameraRight;
		xPositionWorld.xyz += a_normal.y * fScale * u_xCameraUp;
	
		// TODO - not 100% sure about the sign of these yet
		// Normal should definitely be opposite to camera direction
		xNormal = -u_xCameraForwards;
	}
	else
#endif // #ifndef LIGHT_MAP
	{
		xNormal = RotateVectorIntoWorldSpace( a_normal, u_xInstanceOrientation[0].xyz, u_xInstanceOrientation[1].xyz, u_xInstanceOrientation[2].xyz);	
	}
	
	// final out position
	xPositionWorld.xyz += u_xInstancePosition.xyz;
	gl_Position = u_ViewProj * xPositionWorld;

#ifndef LIGHT_MAP	
	vec3 LightColour = SHLookup( vec4( xNormal, 1.0 ), u_xInstanceRGB_SH[0], u_xInstanceRGB_SH[1], u_xInstanceRGB_SH[2] );
	
	LightColour *= u_xAlbedoColour;	// 2.0f factor is in the Albedo colour

	v_VertexColour = vec4(LightColour, 1.0);
#else
	v_VertexColour = vec4(u_xAlbedoColour, 1.0);
#endif

	vec4 xTexCoords = a_texcoord * vec4(1.0/1024.0, 1.0/1024.0, 1.0/16384.0, 1.0/16384.0);

	v_TexCoord = xTexCoords.xy + u_TexAnim;
#ifdef LIGHT_MAP
    vec3 xUVLightMapScale = vec3( u_xInstanceOrientation[0].w, u_xInstanceOrientation[1].w, u_xInstanceOrientation[2].w );
    v_LightMapUV = xTexCoords.zw * xUVLightMapScale.xx + xUVLightMapScale.yz;
#endif

#ifdef ENVIRONMENT_MAP
	mediump vec3 camdir = normalize(xPositionWorld.xyz - u_CameraPosition);
	v_Reflection = reflect(camdir, xNormal);
#endif
	
#ifdef FOG  
	vec4 ViewPos = u_View * xPositionWorld;
    float depth = length(ViewPos.xyz);
  
	float fog_depth = clamp((depth - u_FogNearPlane) * u_FogRcpRange, 0.0, 1.0);
	v_Fog = u_FogOpacity * fog_depth;
#endif// FOG

}